@@ -1,9 +1,9 @@
 package Net::DNS::Text;
 
-#
-# $Id: Text.pm 1698 2018-07-24 15:29:05Z willem $
-#
-our $VERSION = (qw$LastChangedRevision: 1698 $)[1];
+use strict;
+use warnings;
+
+our $VERSION = (qw$Id: Text.pm 1855 2021-11-26 11:33:48Z willem $)[2];
 
 
 =head1 NAME
@@ -14,11 +14,11 @@ Net::DNS::Text - DNS text representation
 
     use Net::DNS::Text;
 
-    $object = new Net::DNS::Text('example');
+    $object = Net::DNS::Text->new('example');
     $string = $object->string;
 
-    $object = decode Net::DNS::Text( \$data, $offset );
-    ( $object, $next ) = decode Net::DNS::Text( \$data, $offset );
+    $object = Net::DNS::Text->decode( \$data, $offset );
+    ( $object, $next ) = Net::DNS::Text->decode( \$data, $offset );
 
     $data = $object->encode;
     $text = $object->value;
@@ -34,8 +34,6 @@ lifetime.
 =cut
 
 
-use strict;
-use warnings;
 use integer;
 use Carp;
 
@@ -54,7 +52,7 @@ use constant UTF8 => scalar eval {	## not UTF-EBCDIC  [see Unicode TR#16 3.6]
 
 =head2 new
 
-    $object = new Net::DNS::Text('example');
+    $object = Net::DNS::Text->new('example');
 
 Creates a text object which encapsulates a single character
 string component of a resource record.
@@ -68,7 +66,7 @@ interpretation.
 
 =cut
 
-my ( %escape, %unescape );		## precalculated ASCII escape tables
+my ( %escape, %escapeUTF8, %unescape );	## precalculated escape tables
 
 sub new {
 	my $self = bless [], shift;
@@ -84,7 +82,7 @@ sub new {
 
 	while ( length $_ > 255 ) {
 		my $chunk = substr( $_, 0, 255 );		# carve into chunks
-		substr( $chunk, -length($1) ) = '' if $chunk =~ /.([\300-\377][\200-\277]*)$/;
+		$chunk =~ s/[\300-\377][\200-\277]*$//;
 		push @$self, $chunk;
 		substr( $_, 0, length $chunk ) = '';
 	}
@@ -96,9 +94,9 @@ sub new {
 
 =head2 decode
 
-    $object = decode Net::DNS::Text( \$buffer, $offset );
+    $object = Net::DNS::Text->decode( \$buffer, $offset );
 
-    ( $object, $next ) = decode Net::DNS::Text( \$buffer, $offset );
+    ( $object, $next ) = Net::DNS::Text->decode( \$buffer, $offset );
 
 Creates a text object which represents the decoded data at the
 indicated offset within the data buffer.
@@ -142,7 +140,7 @@ suitable for inclusion in a DNS packet buffer.
 
 sub encode {
 	my $self = shift;
-	join '', map pack( 'C a*', length $_, $_ ), @$self;
+	return join '', map { pack( 'C a*', length $_, $_ ) } @$self;
 }
 
 
@@ -157,7 +155,7 @@ without the explicit length field.
 
 sub raw {
 	my $self = shift;
-	join '', map pack( 'a*', $_ ), @$self;
+	return join '', map { pack( 'a*', $_ ) } @$self;
 }
 
 
@@ -172,7 +170,7 @@ Character string representation of the text object.
 sub value {
 	return unless defined wantarray;
 	my $self = shift;
-	_decode_utf8( join '', @$self );
+	return _decode_utf8( join '', @$self );
 }
 
 
@@ -180,20 +178,33 @@ sub value {
 
     $string = $text->string;
 
-Conditionally quoted zone file representation of the text object.
+Conditionally quoted RFC1035 zone file representation of the text object.
 
 =cut
 
 sub string {
 	my $self = shift;
 
-	my @s = map split( '', $_ ), @$self;			# escape special and ASCII non-printable
-	my $string = _decode_utf8( join '', map $escape{$_}, @s );
+	my @s = map { split '', $_ } @$self;			# escape special and ASCII non-printable
+	my $s = _decode_utf8( join '', map { $escape{$_} } @s );
+	return $s =~ /[ \t\n\r\f(),;]|^$/ ? qq("$s") : $s;	# quote special characters and empty string
+}
 
-	return $string unless $string =~ /[ \t\n\r\f]|^$|;$/;	# unquoted contiguous
 
-	$string =~ s/\\([^"0-9])/$1/g;				# unescape printable characters except \"
-	join '', '"', $string, '"';				# quoted string
+=head2 unicode
+
+    $string = $text->unicode;
+
+Conditionally quoted Unicode representation of the text object.
+
+=cut
+
+sub unicode {
+	my $self = shift;
+
+	my @s = map { split '', $_ } @$self;			# escape special and non-printable
+	my $s = _decode_utf8( join '', map { $escapeUTF8{$_} } @s );
+	return $s =~ /[ \t\n\r\f();]|^$/ ? qq("$s") : $s;	# quote special characters and empty string
 }
 
 
@@ -213,7 +224,7 @@ sub _decode_utf8 {			## UTF-8 to perl internal encoding
 	[ !"#$%&'()*+,\-./0-9:;<=>?@A-Z\[\\\]^_`a-z{|}~?] unless ASCII;
 
 	my $z = length($_) - length($_);			# pre-5.18 taint workaround
-	ASCII ? substr( ( UTF8 ? $utf8 : $ascii )->decode($_), $z ) : $_;
+	return ASCII ? substr( ( UTF8 ? $utf8 : $ascii )->decode($_), $z ) : $_;
 }
 
 
@@ -226,35 +237,30 @@ sub _encode_utf8 {			## perl internal encoding to UTF-8
 	[\040-\176] unless ASCII;
 
 	my $z = length($_) - length($_);			# pre-5.18 taint workaround
-	ASCII ? substr( ( UTF8 ? $utf8 : $ascii )->encode($_), $z ) : $_;
+	return ASCII ? substr( ( UTF8 ? $utf8 : $ascii )->encode($_), $z ) : $_;
 }
 
 
-%escape = eval {			## precalculated ASCII/UTF-8 escape table
-	my %table;
-	my @C0 = ( 0 .. 31 );					# control characters
-	my @NA = UTF8 ? ( 192, 193, 216 .. 223, 245 .. 255 ) : ( 128 .. 255 );
+%escape = eval {			## precalculated ASCII escape table
+	my %table = map { ( chr($_) => chr($_) ) } ( 0 .. 127 );
 
-	foreach ( 0 .. 255 ) {					# transparent
-		$table{pack( 'C', $_ )} = pack 'C', $_;
-	}
-
-	foreach ( 34, 40, 41, 59 ) {				# character escape	" ( ) ;
-		$table{pack( 'C', $_ )} = pack 'C2', 92, $_;
-	}
-
-	foreach my $n ( @C0, 92, 127, @NA ) {			# numerical escape
+	foreach my $n ( 0 .. 31, 34, 92, 127 .. 255 ) {		# numerical escape
 		my $codepoint = sprintf( '%03u', $n );
 
-		# partial transliteration for non-ASCII character encodings
+		# transliteration for non-ASCII character encodings
 		$codepoint =~ tr [0-9] [\060-\071];
 
-		$table{pack( 'C', $n )} = pack 'C a3', 92, $codepoint;
+		$table{chr($n)} = pack 'C a3', 92, $codepoint;
 	}
 
 	return %table;
 };
 
+%escapeUTF8 = eval {			## precalculated UTF-8 escape table
+	my @octet = UTF8 ? ( 128 .. 191, 194 .. 254 ) : ();
+	return ( %escape, map { ( chr($_) => chr($_) ) } @octet );
+};
+
 
 %unescape = eval {			## precalculated numeric escape table
 	my %table;
@@ -262,12 +268,12 @@ sub _encode_utf8 {			## perl internal encoding to UTF-8
 	foreach my $n ( 0 .. 255 ) {
 		my $key = sprintf( '%03u', $n );
 
-		# partial transliteration for non-ASCII character encodings
+		# transliteration for non-ASCII character encodings
 		$key =~ tr [0-9] [\060-\071];
 
 		$table{$key} = pack 'C', $n;
-		$table{$key} = pack 'C2', 92, $n if $n == 92;	   # escaped escape
 	}
+	$table{"\060\071\062"} = pack 'C2', 92, 92;		# escaped escape
 
 	return %table;
 };
@@ -300,7 +306,7 @@ All rights reserved.
 
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose and without fee is hereby granted, provided
-that the above copyright notice appear in all copies and that both that
+that the original copyright notices appear in all copies and that both
 copyright notice and this permission notice appear in supporting
 documentation, and that the name of the author not be used in advertising
 or publicity pertaining to distribution of the software without specific
